import importlib
import os
import shutil
import tempfile
from typing import Dict, Tuple
from urllib.request import urlopen

import yaml
from tqdm import tqdm

__PACKAGE_DIR__ = os.path.dirname(__file__)


def _get_local_package_dir() -> str:
    """
    Get the path to the local package directory.

    Returns:
        str: Path to the local package directory.
    """
    assert os.path.exists(
        __PACKAGE_DIR__
    ), f"package directory not found: {__PACKAGE_DIR__}"
    return __PACKAGE_DIR__


def _get_config_path(config_path=None) -> str:
    """
    Get the path to the config file.

    Args:
        config_path (_type_, optional): Path to the config file. Defaults to None.

    Raises:
        ValueError: If the config_path is not a valid path.

    Returns:
        str: Path to the config file.
    """
    if config_path is None:
        config_path = os.path.join(__PACKAGE_DIR__, ".config.yaml")
    if not os.path.exists(config_path):
        raise ValueError("Config path {} does not exist!".format(config_path))
    return config_path


def _load_config_file(config_path=None) -> Dict:
    """
    Load the config file.

    Args:
        config_path (_type_, optional): Path to the config file. Defaults to None.

    Returns:
        Dict: The config file as a dictionary.
    """
    yaml_path = _get_config_path(config_path)
    Loader = yaml.FullLoader
    with open(yaml_path, "r") as config_file:
        cfg = yaml.load(config_file, Loader=Loader)

    return cfg


def _get_from_config_file(key: str = None, cfg: Dict = None):
    """
    Get a value from the config file.

    Args:
        key (str, optional): Key to the config file. Defaults to None.
        cfg (Dict, optional): Config file as a dictionary. Defaults to None.

    Returns:
        _type_: _description_
    """
    if isinstance(key, type(None)):
        if isinstance(cfg, type(None)):
            cfg = _load_config_file()
        return cfg
    else:
        if isinstance(cfg, type(None)):
            cfg = _load_config_file()

        key, *next_key = key.split(".")
        next_key = ".".join(next_key) if len(next_key) > 0 else None
        assert key in cfg.keys(), "yaml file does not contains given key"
        cfg = cfg.get(key)
    return _get_from_config_file(key=next_key, cfg=cfg)


def _get_model_dir() -> str:
    """
    Get the path to the model directory.

    Returns:
        str: Path to the model directory.
    """
    model_registry_dir = os.path.join(
        _get_local_package_dir(), _get_from_config_file("PACKAGE.MODEL")
    )
    assert os.path.exists(
        model_registry_dir
    ), f"model directory not found: {model_registry_dir}"
    return model_registry_dir


def _get_model_info(model_name, version) -> str:
    """
    Get the model info.

    Args:
        model_name (_type_): Model name.
        version (_type_): Model version.

    Returns:
        str: Model info.
    """
    # Get arch name and config name from the given model_name
    arch, config, dataset = _parse_saved_model_name(model_name)

    information = _get_from_config_file(f"MODEL.{arch}.{config}.{dataset}.{version}")

    return information


def _get_arch_dir() -> str:
    """
    Get the path to the arch directory.

    Returns:
        str: Path to the arch directory.
    """
    arch_registry_dir = os.path.join(
        _get_local_package_dir(), _get_from_config_file("PACKAGE.ARCH")
    )
    assert os.path.exists(
        arch_registry_dir
    ), f"architecture directory not found: {arch_registry_dir}"
    return arch_registry_dir


def _get_arch_pkg(arch: str):
    """
    Get the architecture package.

    Args:
        arch (str): Architecture name.

    Returns: Package containing the architecture.
    """
    pkg_name = _get_from_config_file("PACKAGE.NAME")
    arch_postfix = _get_from_config_file("PACKAGE.ARCH")
    return importlib.import_module(".".join([pkg_name, arch_postfix, arch]))


def _get_arch_cls(arch: str):
    """
    Get the architecture class.

    Args:
        arch (str): Architecture name.

    Returns: Architecture class.
    """
    arch_pkg = _get_arch_pkg(arch)
    for arch_name in dir(arch_pkg):
        if arch_name.lower() == arch:
            return getattr(arch_pkg, arch_name)


def _get_list_of_dirs(dir_path: str) -> list:
    """
    Get the list of directories in the given directory.

    Args:
        dir_path (str): Directory path.

    Returns:
        list: List of directories.
    """
    return [
        d
        for d in os.listdir(dir_path)
        if os.path.isdir(os.path.join(dir_path, d)) and not d.startswith("_")
    ]


def _parse_saved_model_name(model: str) -> Tuple[str, str, str]:
    """
    Parse the saved model name.

    Args:
        model (str): Saved model name.

    Returns:
        Tuple[str, str, str]: Architecture name, config name, dataset name.
    """
    arch = None
    config = None
    dataset = None
    splits = model.split("_")
    assert len(splits) in [
        3,
    ], f"model must contain 3 under scores (_) but found: {len(splits)}"
    arch, config, dataset = splits
    return arch, config, dataset


def _download_file_from_url(url, dst, progress=True):
    """
    Download a file from a url.

    Args:
        url (_type_): URL to download the file from.
        dst (_type_): Destination path to save the file to.
        progress (bool, optional): Show progress bar. Defaults to True.
    """

    file_size = None
    u = urlopen(url)
    meta = u.info()
    if hasattr(meta, "getheaders"):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)

    try:
        with tqdm(
            total=file_size,
            disable=not progress,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                pbar.update(len(buffer))

        f.close()

        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)



from typing import Dict, List

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image, ImageDraw, ImageFont


def configure_batch(
    batch: List[torch.Tensor],
    target_size: int,
    adaptive_batch: bool = False,
) -> torch.Tensor:
    """
    Configure batch for the required size

    Args:
        batch (List[torch.Tensor]): List of torch.Tensor(C, H, W)
        target_size (int): Max dimension of the target image
        adaptive_batch (bool, optional): If true, the batch will be adaptive (target_size is the max dimension of the each image), otherwise it will use given `target_size`. Defaults to False.

    Returns:
        torch.Tensor: batched inputs as B x C x target_size x target_size
    """
    # pylint: disable=no-member

    for i, img in enumerate(batch):

        # Check adaptive batch is given
        if adaptive_batch:
            # Get max dimension of the image
            target_size: int = max(img.size(1), img.size(2))

        # Apply interpolation to the image
        img_h: int = img.size(-2)
        img_w: int = img.size(-1)

        scale_factor: float = min(target_size / img_h, target_size / img_w)

        img = F.interpolate(
            img.unsqueeze(0),
            scale_factor=scale_factor,
            mode="bilinear",
            recompute_scale_factor=False,
            align_corners=False,
        )

        new_h: int = img.size(-2)
        new_w: int = img.size(-1)

        # Apply padding to the image
        pad_left = (target_size - new_w) // 2
        pad_right = pad_left + (target_size - new_w) % 2

        pad_top = (target_size - new_h) // 2
        pad_bottom = pad_top + (target_size - new_h) % 2

        img = F.pad(img, (pad_left, pad_right, pad_top, pad_bottom), value=0)

        batch[i] = img

    batch = torch.cat(batch, dim=0).contiguous()

    return batch


def convert_json(
    preds: List[torch.Tensor],
    label: List[str],
) -> List[Dict]:
    """
    Convert the predictions to a json format

    Args:
        preds (List[torch.Tensor]): List of torch.Tensor
        label (List[str]): List of labels

    Returns:
        List[Dict]: List of dictionaries with the label as key and the score as value
    """
    outputs = []
    for scores in preds.cpu().numpy().tolist():
        output = {
            label: round(score, 2)
            for score, label in zip(
                scores,
                label,
            )
        }
        outputs.append(
            dict(sorted(output.items(), key=lambda item: item[1], reverse=True))
        )

    return outputs


def visualize(img: np.ndarray, preds: List[Dict]) -> Image:
    """
    Virtualize the image with the predictions

    Args:
        img (np.ndarray): 3 channel np.ndarray
        preds (List[Dict]): Predictions. Each prediction is a dictionary with the label as key and the score as value

    Returns:
        Image: 3 channel PIL Image that will be shown on screen
    """
    pil_img = Image.fromarray(img)
    pil_img = pil_img.resize((512, 512), Image.ANTIALIAS)
    font = ImageFont.load_default()

    for pred in preds:
        max_key = max(pred, key=pred.get)
        model_info = f" {max_key} : {pred[max_key]}"

    text_width, text_height = font.getsize(model_info)
    margin = np.ceil(0.05 * text_height)

    ImageDraw.Draw(pil_img).text(
        (margin, text_height - margin),
        model_info,
        fill="white",
        font=font,
    )

    return pil_img



from typing import Callable, Optional

import torch
from torch import nn


class ConvNormActivation(torch.nn.Sequential):
    """
    Configurable block used for Convolution-Normalzation-Activation blocks.

    Args:
        in_channels (int): Number of channels in the input image
        out_channels (int): Number of channels produced by the Convolution-Normalzation-Activation block
        kernel_size: (int, optional): Size of the convolving kernel. Default: 3
        stride (int, optional): Stride of the convolution. Default: 1
        padding (int, tuple or str, optional): Padding added to all four sides of the input. Default: None, in wich case it will calculated as ``padding = (kernel_size - 1) // 2 * dilation``
        groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
        norm_layer (Callable[..., torch.nn.Module], optional): Norm layer that will be stacked on top of the convolutiuon layer. If ``None`` this layer wont be used. Default: ``torch.nn.BatchNorm2d``
        activation_layer (Callable[..., torch.nn.Module], optinal): Activation function which will be stacked on top of the normalization layer (if not None), otherwise on top of the conv layer. If ``None`` this layer wont be used. Default: ``torch.nn.ReLU``
        dilation (int): Spacing between kernel elements. Default: 1
        inplace (bool): Parameter for the activation layer, which can optionally do the operation in-place. Default ``True``
        bias (bool, optional): Whether to use bias in the convolution layer. By default, biases are included if ``norm_layer is None``.

    This implementation is taken from torchvision repositories.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: Optional[int] = None,
        groups: int = 1,
        norm_layer: Optional[Callable[..., torch.nn.Module]] = torch.nn.BatchNorm2d,
        activation_layer: Optional[Callable[..., torch.nn.Module]] = torch.nn.ReLU,
        dilation: int = 1,
        inplace: Optional[bool] = True,
        bias: Optional[bool] = None,
    ) -> None:
        if padding is None:
            padding = (kernel_size - 1) // 2 * dilation
        if bias is None:
            bias = norm_layer is None
        layers = [
            torch.nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size,
                stride,
                padding,
                dilation=dilation,
                groups=groups,
                bias=bias,
            )
        ]
        if norm_layer is not None:
            layers.append(norm_layer(out_channels))
        if activation_layer is not None:
            params = {} if inplace is None else {"inplace": inplace}
            layers.append(activation_layer(**params))
        super().__init__(*layers)
        self.out_channels = out_channels


class ConvBNReLU(nn.Sequential):
    def __init__(
        self,
        in_planes,
        out_planes,
        kernel_size=3,
        stride=1,
        groups=1,
        norm_layer=None,
    ):
        padding = (kernel_size - 1) // 2
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        super(ConvBNReLU, self).__init__(
            nn.Conv2d(
                in_planes,
                out_planes,
                kernel_size,
                stride,
                padding,
                groups=groups,
                bias=False,
            ),
            norm_layer(out_planes),
            nn.ReLU6(inplace=True),
        )


class InvertedResidual(nn.Module):
    def __init__(
        self,
        inp,
        oup,
        stride,
        expand_ratio,
        norm_layer=None,
    ):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(
                ConvBNReLU(inp, hidden_dim, kernel_size=1, norm_layer=norm_layer)
            )
        layers.extend(
            [
                # dw
                ConvBNReLU(
                    hidden_dim,
                    hidden_dim,
                    stride=stride,
                    groups=hidden_dim,
                    norm_layer=norm_layer,
                ),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                norm_layer(oup),
            ]
        )
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


def _make_divisible(
    v,
    divisor,
    min_value=None,
):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    This implementation is taken from torchvision repositories.
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v



import errno
import os
import sys
from typing import List

from ..core import (
    _download_file_from_url,
    _get_arch_cls,
    _get_arch_dir,
    _get_from_config_file,
    _get_list_of_dirs,
    _get_model_dir,
    _get_model_info,
)

# APIs for Model


def available_models() -> list:
    """
    List of available models.

    Returns:
        list: List of available models.
    """
    # pylint: disable=E1136

    model_list = []
    model_config = _get_from_config_file("MODEL")
    for arch in model_config.keys():
        for config in model_config[arch].keys():
            for dataset in model_config[arch][config].keys():
                model_list.append(f"{arch}_{config}_{dataset}")
    return sorted(model_list)


def get_model_versions(model: str) -> list:
    """
    Get list of available versions for the given model.

    Args:
        model (str): Model name.

    Returns:
        list: List of available versions.
    """
    model_dir = _get_model_dir()
    assert model in available_models(), f"given model: {model} not in the registry"
    model_path = os.path.join(model_dir, model)

    versions = []
    for version in _get_list_of_dirs(model_path):
        versions.append(version)

    return sorted(versions)


def get_model_latest_version(model: str) -> int:
    """
    Get latest version of the given model.

    Args:
        model (str): Model name.

    Returns:
        int: Latest version.
    """
    return max(get_model_versions(model))


def get_saved_model(model: str, version: str):
    """
    Get saved model.

    Args:
        model (str): Model name.
        version (str): Model version.
    """
    model_dir = os.path.join(_get_model_dir(), model, str(version))

    info_dict = _get_model_info(model, version)
    url = info_dict.get("url")

    try:
        os.makedirs(model_dir)
    except OSError as os_error:
        if os_error.errno != errno.EEXIST:
            raise
        # time.sleep might help here

    cached_file = os.path.join(model_dir, model + ".pth")

    if not os.path.exists(cached_file):
        sys.stderr.write(f'Downloading: "{url}" to {cached_file}\n')
        _download_file_from_url(url, cached_file, progress=True)

    # ToDo: Make it for zip file


# APIs for Architecture


def available_archs() -> List[str]:
    """
    List of available architectures.

    Returns:
        List[str]: List of available architectures.
    """
    archs_dir = _get_arch_dir()
    dirs = _get_list_of_dirs(archs_dir)
    return sorted(dirs)


def get_arch_configs(arch: str) -> List[str]:
    """
    List of available configurations for the given architecture.

    Args:
        arch (str): Architecture name.

    Returns:
        List[str]: List of available configurations.
    """
    cls = _get_arch_cls(arch)
    configs = getattr(cls, "__CONFIGS__", {})
    return sorted(list(configs.keys()))



from typing import Any, Callable, List, Optional

import torch
from torch import nn

from .layers import ConvBNReLU, ConvNormActivation, InvertedResidual, _make_divisible


class MobileNet_V2(nn.Module):
    # pylint: disable=no-member
    def __init__(
        self,
        num_classes: int = 1000,
        width_mult: float = 1.0,
        inverted_residual_setting: Optional[List[List[int]]] = None,
        round_nearest: int = 8,
        block: Optional[Callable[..., nn.Module]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        dropout: float = 0.2,
    ) -> None:
        """
        MobileNet V2 main class
        Args:
            num_classes (int): Number of classes
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for mobilenet
            norm_layer: Module specifying the normalization layer to use
            dropout (float): The droupout probability

        This implementation is taken from torchvision repositories.
        `torchvision.models.mobilenetv2`
        Here is a link for it: `github.com/pytorch/vision`
        """
        super().__init__()

        if block is None:
            block = InvertedResidual

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        input_channel = 32
        last_channel = 1280

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        # only check the first element, assuming user knows t,c,n,s are required
        if (
            len(inverted_residual_setting) == 0
            or len(inverted_residual_setting[0]) != 4
        ):
            raise ValueError(
                f"inverted_residual_setting should be non-empty or a 4-element list, got {inverted_residual_setting}"
            )

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(
            last_channel * max(1.0, width_mult), round_nearest
        )
        features: List[nn.Module] = [
            ConvNormActivation(
                3,
                input_channel,
                stride=2,
                norm_layer=norm_layer,
                activation_layer=nn.ReLU6,
            )
        ]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(
                    block(
                        input_channel,
                        output_channel,
                        stride,
                        expand_ratio=t,
                        norm_layer=norm_layer,
                    )
                )
                input_channel = output_channel
        # building last several layers
        features.append(
            ConvNormActivation(
                input_channel,
                self.last_channel,
                kernel_size=1,
                norm_layer=norm_layer,
                activation_layer=nn.ReLU6,
            )
        )
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        # building classifier
        self.classifier = nn.Sequential(
            nn.Dropout(p=dropout),
            nn.Linear(self.last_channel, num_classes),
        )

        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.features(x)
        x = nn.functional.adaptive_avg_pool2d(x, (1, 1))
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x



from typing import List, Dict, Tuple
import copy

import torch
from torch.utils.data import Dataset
import numpy as np
import imageio


class BaseDataset(Dataset):
    # pylint: disable=no-member
    # pylint: disable=not-callable

    def __init__(
        self,
        ids: List[str],
        targets: List[Dict],
        transforms=None,
        **kwargs,
    ):
        super().__init__()
        assert isinstance(ids, list), "given `ids` must be list"
        assert isinstance(targets, list), "given `targets must be list"
        assert len(ids) == len(targets), "lenght of both lists must be equal"

        self.ids = ids
        self.targets = targets
        self.transforms = transforms

        # set given kwargs to the dataset
        for key, value in kwargs.items():
            if hasattr(self, key):
                continue
            setattr(self, key, value)

    def __getitem__(self, idx: int) -> Tuple:
        img = self._load_image(self.ids[idx])
        img = torch.from_numpy(img).permute(2, 0, 1)
        targets = copy.deepcopy(self.targets[idx])

        # apply transforms
        if self.transforms:
            img, targets = self.transforms(img, targets)
        targets = torch.tensor(targets, dtype=torch.long)

        return (img, targets)

    def __len__(self) -> int:
        return len(self.ids)

    @staticmethod
    def _load_image(img_file_path: str) -> np.ndarray:

        img = imageio.imread(img_file_path)

        if not img.flags["C_CONTIGUOUS"]:
            # if img is not contiguous than fix it
            img = np.ascontiguousarray(img, dtype=img.dtype)

        if len(img.shape) == 4:
            # found RGBA, converting to => RGB
            img = img[:, :, :3]
        elif len(img.shape) == 2:
            # found GRAYSCALE, converting to => RGB
            img = np.stack([img, img, img], axis=-1)
        else:
            img = img[:, :, :3]

        return np.array(img, dtype=np.uint8)



import os
from typing import Dict, List

import torch
import torch.nn as nn
import torchvision

from .blocks.mobilenetv2 import MobileNet_V2


class MobileNetV2(nn.Module):
    """
    Implementation of MobileNetV2: Inverted Residuals and Linear Bottlenecks

    The network is the one described in arxiv.org/abs/1801.04381v4 .
    """

    # pylint: disable=no-member

    __CONFIGS__ = {
        "default": {
            "input": {
                "input_size": 224,
                "normalized_input": True,
                "mean": [0.485, 0.456, 0.406],
                "std": [0.229, 0.224, 0.225],
            },
            "model": {
                "width_mult": 1.0,
                "inverted_residual_setting": None,
                "round_nearest": 8,
                "block": None,
                "norm_layer": None,
                "dropout": 0.2,
                "pretrained": False,
            },
        },
    }

    def __init__(
        self,
        config: Dict,
        labels: List[str] = None,
    ):
        super().__init__()
        self.config = config
        self.labels = labels
        self.num_classes = len(self.labels)

        self.backbone = MobileNet_V2(
            num_classes=self.num_classes,
            width_mult=self.config["model"]["width_mult"],
            inverted_residual_setting=self.config["model"]["inverted_residual_setting"],
            round_nearest=self.config["model"]["round_nearest"],
            block=self.config["model"]["block"],
            norm_layer=self.config["model"]["norm_layer"],
            dropout=self.config["model"]["dropout"],
        )

        if self.config["model"]["pretrained"]:
            self.backbone.load_state_dict(
                torchvision.models.mobilenet_v2(pretrained=True).state_dict(),
                strict=False,
            )

    def forward(self, inputs):
        """
        Forward pass of the model.
        """
        return self.backbone(inputs)

    def logits_to_preds(self, logits: List[torch.Tensor]):
        """
        Convert logits to predictions.
        """
        return torch.softmax(logits, axis=-1)

    @classmethod
    def build(
        cls,
        config: str = "",
        labels: List[str] = None,
        **kwargs,
    ) -> nn.Module:
        """
        Build the model with random weights.

        Args:
            config (str, optional): Configuration name. Defaults to "".
            labels (List[str], optional): List of labels. Defaults to None.

        Returns:
            nn.Module: Model with random weights.
        """
        # return model with random weight initialization
        labels = ["cls1", "cls2"] if labels is None else labels

        return cls(
            config=MobileNetV2.__CONFIGS__[config],
            labels=labels,
            **kwargs,
        )

    @classmethod
    def from_pretrained(
        cls,
        model_path: str,
        config: str,
        *args,
        **kwargs,
    ) -> nn.Module:
        """
        Load a model from a pre-trained model.

        Args:
            model_path (str): Path to the pre-trained model
            config (str): Configuration of the model

        Returns:
            nn.Module: Model with pretrained weights
        """

        *_, full_model_name, _ = model_path.split(os.path.sep)

        s_dict = torch.load(os.path.join(model_path, f"{full_model_name}.pth"))

        model = cls(
            config=MobileNetV2.__CONFIGS__[config],
            labels=s_dict["labels"],
            *args,
            **kwargs,
        )

        model.load_state_dict(s_dict["state_dict"], strict=True)

        return model

    def compute_loss(
        self,
        logits: List[torch.Tensor],
        targets: List,
        hparams: Dict,
    ):
        """
        Compute the loss for the model.

        Args:
            logits (List[torch.Tensor]): _description_
            targets (List): _description_
            hparams (Dict, optional): _description_. Defaults to {}.

        Raises:
            ValueError: Unknown criterion

        Returns:  Loss
        """
        if hparams.get("criterion", "cross_entropy") == "cross_entropy":
            loss_fcn = nn.CrossEntropyLoss()
        else:
            raise ValueError("Unknown criterion")

        return {"loss": loss_fcn(logits, targets)}

    def configure_optimizers(self, hparams: Dict):
        """
        Configure optimizers for the model.

        Args:
            hparams (Dict): Hyperparameters

        Raises:
            ValueError: Unknown optimizer
            ValueError: Unknown scheduler

        Returns: optimizers and scheduler
        """
        hparams_optimizer = hparams.get("optimizer", "sgd")
        if hparams_optimizer == "sgd":
            optimizer = torch.optim.SGD(
                self.parameters(),
                lr=hparams.get("learning_rate", 1e-1),
                momentum=hparams.get("momentum", 0.9),
                weight_decay=hparams.get("weight_decay", 1e-5),
            )
        elif hparams_optimizer == "adam":
            optimizer = torch.optim.Adam(
                self.parameters(),
                lr=hparams.get("learning_rate", 1e-1),
                betas=hparams.get("betas", (0.9, 0.999)),
                eps=hparams.get("eps", 1e-08),
                weight_decay=hparams.get("weight_decay", 1e-5),
            )
        else:
            raise ValueError("Unknown optimizer")

        hparams_scheduler = hparams.get("scheduler", "steplr")
        if hparams_scheduler == "steplr":
            scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer,
                step_size=hparams.get("step_size", 4),
                gamma=hparams.get("gamma", 0.5),
            )
        elif hparams_scheduler == "multisteplr":
            scheduler = torch.optim.lr_scheduler.MultiStepLR(
                optimizer,
                gamma=hparams.get("gamma", 0.5),
                milestones=hparams.get("milestones", [500000, 1000000, 1500000]),
            )
        else:
            raise ValueError("Unknown scheduler")

        return [optimizer], [scheduler]



import os
from typing import List, Tuple

import numpy as np
from sklearn import preprocessing
from sklearn.model_selection import train_test_split
from torchvision.datasets.utils import extract_archive
from .base import BaseDataset
from ..core import _download_file_from_url


class EuroSAT(BaseDataset):
    """
    Eurosat is a dataset and deep learning benchmark for land use and land cover classification. The dataset is based on Sentinel-2 satellite images covering 13 spectral bands and consisting out of 10 classes with in total 27,000 labeled and geo-referenced images.
    """

    __phases__ = ("val", "train")

    def __init__(
        self,
        root_dir: str = None,
        phase: str = "val",
        transforms=None,
        **kwargs,
    ):
        if root_dir is None:
            root_dir = "TODO"

        self.root_dir = root_dir
        self.phase = phase
        self.download()

        ids, targets = self._split_dataset(phase)
        super().__init__(ids, targets, transforms=transforms, **kwargs)

    @property
    def num_classes(self) -> int:
        return len(self.__classes)

    @property
    def classes(self) -> List[str]:
        return self.__classes

    def name_to_id(self, name: str) -> int:
        return self.__classes.index(name)

    def id_to_name(self, idx: int) -> str:
        return self.__classes[idx]

    def _split_dataset(self, phase: str) -> Tuple:
        labels = []
        filenames = []

        data_dir = os.path.join(self.root_dir)

        for item in os.listdir(data_dir):
            f = os.path.join(data_dir, item)
            if os.path.isfile(f):
                continue
            for subitem in os.listdir(f):
                sub_f = os.path.join(f, subitem)
                filenames.append(sub_f)
                labels.append(item)

        filenames = np.asarray(filenames)
        labels = np.asarray(labels)

        labels = labels[filenames.argsort()]
        filenames = filenames[filenames.argsort()]

        # convert to integer labels
        label_encoder = preprocessing.LabelEncoder()
        label_encoder.fit(np.sort(np.unique(labels)))
        labels = label_encoder.transform(labels)
        labels = np.asarray(labels)

        # remember label encoding
        self.__classes = list(label_encoder.classes_)

        # split into a is_train and test set as provided data is not presplit
        x_train, x_test, y_train, y_test = train_test_split(
            filenames,
            labels,
            test_size=0.2,
            random_state=1,
            stratify=labels,
        )
        if phase == "train":
            return x_train.tolist(), y_train.tolist()
        else:
            return x_test.tolist(), y_test.tolist()

    def _check_exists(self) -> bool:
        """
        Check the Root directory is exists
        """
        return os.path.exists(self.root_dir)

    def download(self) -> None:
        """
        Download the dataset from the internet
        """

        if self._check_exists():
            return

        os.makedirs(self.root_dir, exist_ok=True)
        _download_file_from_url(
            "https://drive.google.com/u/0/uc?id=1_QZVrVVmybvY8_pJjCdnev9THLvS3739&export=download&confirm=t",
            os.path.join(self.root_dir, "eurosat.zip"),
        )
        extract_archive(
            os.path.join(self.root_dir, "eurosat.zip"),
            self.root_dir,
            remove_finished=True,
        )


if __name__ == "__main__":
    data = EuroSAT("satellighte/datas/eurosat")
    print(data[0])



from .api import *
from .core import *
from .datasets import *
from .module import Classifier
from .utils import *
from .version import __version__
from .__about__ import __description__, __license__, __license_url__, __author__



import satellighte.version as pkg
import fire


def write_version_file(major: int, minor: int, patch: int):
    """
    Write the version file.

    Args:
        major (int): Major version.
        minor (int): Minor version.
        patch (int): Patch version.
    """
    # pylint: disable=unspecified-encoding

    version = f'__version__ = "{major}.{minor}.{patch}"'
    with open(pkg.__file__, "w") as version_file:
        version_file.write(version)


def inc_patch():
    """
    Increment the patch version.
    """
    version = pkg.__version__
    major, minor, patch = version.split(".")
    write_version_file(major, minor, int(patch) + 1)


def inc_minor():
    """
    Increment the minor version.
    """
    version = pkg.__version__
    major, minor, _ = version.split(".")
    write_version_file(major, int(minor) + 1, 0)


def inc_major():
    """
    Increment the major version.
    """
    version = pkg.__version__
    major, _, _ = version.split(".")
    write_version_file(int(major) + 1, 0, 0)


if __name__ == "__main__":
    fire.Fire(
        {
            "get_version": pkg.__version__,
            "inc-patch": inc_patch,
            "inc-minor": inc_minor,
            "inc-major": inc_major,
        }
    )



import os
from importlib.util import module_from_spec, spec_from_file_location

import setuptools

_PATH_ROOT = os.path.dirname(__file__)


def _load_py_module(fname, pkg="satellighte"):
    spec = spec_from_file_location(
        os.path.join(pkg, fname), os.path.join(_PATH_ROOT, pkg, fname)
    )
    py_module = module_from_spec(spec)
    spec.loader.exec_module(py_module)
    return py_module


about = _load_py_module("__about__.py")

test_require = [
    "pytest>=6.0.0",
    "pytest-pylint",
    "pytest-cov",
    "pylint",
    "black",
]

deploy_require = [
    "pytest>=6.0.0",
    "pytest-pylint",
    "pytest-cov",
    "pylint",
    "black",
]

docs_require = [
    "sphinxemoji",
    "sphinx_rtd_theme",
    "recommonmark",
    "sphinx_markdown_tables",
    "sphinxcontrib-napoleon",
    "sphinx-autodoc-typehints",
    "nbsphinx",
    "furo",
]

extras_require = {
    "test": test_require,
    "docs": docs_require,
    "deploy": about.__requirements__ + deploy_require,
    "all": about.__requirements__ + test_require + deploy_require,
}
setuptools.setup(
    name=about.__pkg_name__,
    version=about.__version__,
    description=about.__description__,
    author=about.__author__,
    author_email=about.__author_email__,
    url=about.__homepage__,
    license=about.__license__,
    packages=setuptools.find_packages(),
    include_package_data=True,
    long_description=about.__long_description__,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    extras_require=extras_require,
    setup_requires=[],
    install_requires=about.__requirements__,
    classifiers=[
        # Indicate who your project is intended for
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import os
from typing import Dict, List, Union

import numpy as np
import pytorch_lightning as pl
import torch

from . import api
from .core import _get_arch_cls, _get_model_dir, _parse_saved_model_name
from .utils import configure_batch, convert_json
import torchmetrics as tm


class Classifier(pl.LightningModule):
    """Generic pl.LightningModule definition for image classification"""

    # pylint: disable=no-member
    # pylint: disable=not-callable

    def __init__(
        self,
        model: torch.nn.Module,
        hparams: Dict = None,
    ):
        super().__init__()
        self.model = model
        self.__metrics = {}

        self.save_hyperparameters(hparams)
        self.configure_preprocess()

    @property
    def input_size(self):
        return self.model.config.get("input").get("input_size")

    @property
    def labels(self):
        return self.model.labels

    # WARNING: This function should only be used during training. not inference
    def forward(
        self,
        batch: torch.Tensor,
    ) -> torch.Tensor:
        """
        Forward pass of the model.

        Args:
            batch (torch.Tensor): Batch of tensors of shape (B x C x H x W).

        Returns:
            torch.Tensor: Prediction of the model.
        """

        # Apply preprocess with the help of registered buffer
        batch = ((batch / self.normalizer) - self.mean) / self.std

        with torch.no_grad():
            # Get logits from the model
            logits = self.model.forward(batch)

        # Apply postprocess for the logits that are returned from model and get predictions
        preds = self.model.logits_to_preds(logits)

        return preds

    @torch.jit.unused
    def predict(
        self,
        data: Union[np.ndarray, List],
        target_size: int = None,
    ):
        """
        Perform image classification using given image or images.

        Args:
            data (Union[np.ndarray, List]): Numpy array or list of numpy arrays. In the form of RGB.
            target_size (int, optional): If it is not None, the image will be resized to the target size. Defaults to None.

        Returns:
            [type]: [description]
        """

        # Converts given image or list of images to list of tensors
        batch = self.to_tensor(data)

        # Override target_size if input_size is given and target_size is None
        if self.input_size and (target_size is None):
            target_size = self.input_size

        # Configure batch for the required size
        batch = configure_batch(
            batch,
            target_size=target_size,
            adaptive_batch=target_size is None,
        )

        # Get predictions from the model
        preds = self.forward(batch)

        # Convert predictions to json format
        json_preds = convert_json(preds, self.labels)
        return json_preds

    @classmethod
    def build(
        cls,
        arch: str,
        config: str = None,
        hparams: Dict = {},
        **kwargs,
    ) -> pl.LightningModule:
        """
        Build the model with given architecture and configuration.

        Args:
            arch (str): Model architecture name.
            config (str, optional): Model configuration. Defaults to None.
            hparams (Dict, optional): Hyperparameters. Defaults to {}.

        Returns:
            pl.LightningModule: Model instance with randomly initialized weights.
        """

        model = cls.build_arch(arch, config=config, **kwargs)
        return cls(model, hparams=hparams)

    @classmethod
    def build_arch(
        cls,
        arch: str,
        config: str = None,
        **kwargs,
    ) -> torch.nn.Module:
        """
        Build the architecture model with given configuration.

        Args:
            arch (str): Model architecture name.
            config (str, optional): Model configuration. Defaults to None.

        Returns:
            torch.nn.Module: Architecture model instance with randomly initialized weights.
        """

        arch_cls = _get_arch_cls(arch)
        return arch_cls.build(config=config, **kwargs)

    @classmethod
    def from_pretrained(
        cls,
        model_name: str,
        version: int = None,
        hparams: Dict = {},
    ) -> pl.LightningModule:
        """
        [summary]

        Args:
            model_name (str): Model name in the format of {arch}_{config}_{dataset}
            version (int, optional): Model version. Defaults to None.
            hparams (Dict, optional): Hyperparameters. Defaults to {}.

        Returns:
            pl.LightningModule: Model instance.
        """

        model = cls.from_pretrained_arch(model_name, version=version)
        return cls(model, hparams=hparams)

    @classmethod
    def from_pretrained_arch(
        cls,
        model_name: str,
        version: int = None,
    ) -> torch.nn.Module:
        """
        Get pretrained arch model from the model name.

        Args:
            model_name (str): Model name in the format of {arch}_{config}_{dataset}
            version (int, optional): Model version. Defaults to None.

        Returns:
            torch.nn.Module: Architecture model instance.
        """

        # Check if version is not given then get the latest version
        if not version:
            version = api.get_model_latest_version(model_name)

        # Get arch name and config name from the given model_name
        arch, config, _ = _parse_saved_model_name(model_name)

        # Get arch class
        arch_cls = _get_arch_cls(arch)

        api.get_saved_model(model_name, version)

        # Get pretrained model pat
        model_path = os.path.join(_get_model_dir(), model_name, version)

        return arch_cls.from_pretrained(model_path, config=config)

    def training_step(self, batch, batch_idx):
        batch, targets = batch

        # Apply preprocess with the help of registered buffer
        batch = ((batch / self.normalizer) - self.mean) / self.std

        # Get logits from the model
        logits = self.model.forward(batch)

        # Compute loss
        loss = self.model.compute_loss(
            logits,
            targets,
            hparams=self.hparams,
        )

        return loss

    def training_epoch_end(self, outputs):
        losses = {}
        for output in outputs:
            if isinstance(output, dict):
                for k, v in output.items():
                    if k not in losses:
                        losses[k] = []
                    losses[k].append(v)
            else:
                if "loss" not in losses:
                    losses["loss"] = []
                losses["loss"].append(output)

        for name, loss in losses.items():
            self.log("{}/training".format(name), sum(loss) / len(loss))

    def on_validation_epoch_start(self):
        for metric in self.__metrics.values():
            metric.reset()

    def validation_step(self, batch, batch_idx):
        batch, targets = batch

        # Apply preprocess with the help of registered buffer
        batch = ((batch / self.normalizer) - self.mean) / self.std

        with torch.no_grad():
            # Get logits from the model
            logits = self.model.forward(batch)

            # Compute loss
            loss = self.model.compute_loss(
                logits,
                targets,
                hparams=self.hparams,
            )

        # Apply postprocess for the logits that are returned from model and get predictions
        preds = self.model.logits_to_preds(logits)

        for metric in self.__metrics.values():
            metric.update(preds.cpu(), targets.cpu())

        return loss

    def validation_epoch_end(self, outputs):
        losses = {}
        for output in outputs:
            if isinstance(output, dict):
                for k, v in output.items():
                    if k not in losses:
                        losses[k] = []
                    losses[k].append(v)
            else:
                if "loss" not in losses:
                    losses["loss"] = []
                losses["loss"].append(output)

        for name, loss in losses.items():
            self.log("{}/validation".format(name), sum(loss) / len(loss))

        for name, metric in self.__metrics.items():
            self.log(
                "metrics/{}".format(name),
                metric.compute(),
                prog_bar=True,
            )

    def on_test_epoch_start(self):
        for metric in self.__metrics.values():
            metric.reset()

    def test_step(self, batch, batch_idx):
        batch, targets = batch

        # Apply preprocess with the help of registered buffer
        batch = ((batch / self.normalizer) - self.mean) / self.std

        with torch.no_grad():
            # Get logits from the model
            logits = self.model.forward(batch)

            # Compute loss
            loss = self.model.compute_loss(
                logits,
                targets,
                hparams=self.hparams,
            )

        # Apply postprocess for the logits that are returned from model and get predictions
        preds = self.model.logits_to_preds(logits)

        for metric in self.__metrics.values():
            metric.update(preds.cpu(), targets.cpu())

    def test_epoch_end(self, outputs):
        metric_results = {}
        for name, metric in self.__metrics.items():
            metric_results[name] = metric.compute()
        return metric_results

    def configure_optimizers(self):
        return self.model.configure_optimizers(hparams=self.hparams)

    def configure_preprocess(self):
        """
        Configure preprocess for the model.
        """

        # Get information from config of model
        normalized_input = self.model.config.get("input", {}).get(
            "normalized_input", True
        )
        mean = self.model.config.get("input", {}).get("mean", 0.0)
        std = self.model.config.get("input", {}).get("std", 1.0)

        # Check dimension of std and mean
        if isinstance(mean, list):
            assert len(mean) == 3, "mean dimension must be 3 not {}".format(len(mean))
            mean = [float(m) for m in mean]
        else:
            mean = [float(mean) for _ in range(3)]

        if isinstance(std, list):
            assert len(std) == 3, "std dimension must be 3 not {}".format(len(std))
            std = [float(m) for m in std]
        else:
            std = [float(std) for _ in range(3)]

        # Register the tensors as a buffer
        # Now we can access self.normalizer anywhere in the module
        self.register_buffer(
            "normalizer",
            torch.tensor(255.0) if normalized_input else torch.tensor(1.0),
            persistent=False,
        )

        # Now we can access self.mean anywhere in the module
        self.register_buffer(
            "mean",
            torch.tensor(mean).view(-1, 1, 1).contiguous(),
            persistent=False,
        )

        # Now we can access self.std anywhere in the module
        self.register_buffer(
            "std",
            torch.tensor(std).view(-1, 1, 1).contiguous(),
            persistent=False,
        )

    def add_metric(self, name: str, metric: tm.Metric):
        """Adds given metric with name key

        Args:
                name (str): name of the metric
                metric (tm.Metric): Metric object
        """
        # TODO add warnings if override happens
        self.__metrics[name] = metric

    def get_metrics(self) -> Dict[str, tm.Metric]:
        """Return metrics defined in the `FaceDetector` instance

        Returns:
                Dict[str, tm.Metric]: defined model metrics with names
        """
        return {k: v for k, v in self.__metrics.items()}

    def to_tensor(self, images: Union[np.ndarray, List]) -> List[torch.Tensor]:
        """Converts given image or list of images to list of tensors
        Args:
                images (Union[np.ndarray, List]): RGB image or list of RGB images
        Returns:
                List[torch.Tensor]: list of torch.Tensor(C x H x W)

        This method is taken from fastface repositories.
        `fastface.module.to_tensor`
        Here is a link for it: `github.com/borhanMorphy/fastface`
        """
        assert isinstance(
            images, (list, np.ndarray)
        ), "give images must be eather list of numpy arrays or numpy array"

        if isinstance(images, np.ndarray):
            images = [images]

        batch: List[torch.Tensor] = []

        for img in images:
            assert (
                len(img.shape) == 3
            ), "image shape must be channel, height\
                , with length of 3 but found {}".format(
                len(img.shape)
            )
            assert (
                img.shape[2] == 3
            ), "channel size of the image must be 3 but found {}".format(img.shape[2])

            batch.append(
                # h,w,c => c,h,w
                torch.tensor(img, dtype=self.dtype, device=self.device).permute(2, 0, 1)
            )

        return batch



import time
from typing import List

from satellighte.version import __version__ as pkg_version


def _load_requirements(
    file_name: str = "requirements.txt", comment_char: str = "#"
) -> List[str]:
    """
    Load requirements from a text file.

    Args:
        file_name (str, optional): File name. Defaults to "requirements.txt".
        comment_char (str, optional): Disregard lines starting with this character. Defaults to "#".

    Returns:
        List[str]: List of requirements.
    """
    # Open the file
    with open(file_name, "r", encoding="utf-8") as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for line in lines:

        # Disregard comments
        if comment_char in line:
            line = line[: line.index(comment_char)].strip()

        # Disregard http or @http lines
        if line.startswith("http") or "@http" in line:
            continue

        # Add the line to the list
        if line:
            reqs.append(line)
    return reqs


def _load_readme(file_name: str = "README.md") -> str:
    """
    Load readme from a text file.

    Args:
        file_name (str, optional): File name that contains the readme. Defaults to "README.md".

    Returns:
        str: Readme text.
    """
    text = open(file_name, "r", encoding="utf-8").read()

    return text


_this_year = time.strftime("%Y")
__author__ = "Oguzcan Turan"
__author_email__ = "can.turan.10@gmail.com"
__copyright__ = f"Copyright (c) 2022-{_this_year}, {__author__}."
__description__ = (
    "PyTorch Lightning Implementations of Recent Satellite Image Classification !"
)
__homepage__ = "https://github.com/canturan10/satellighte"
__license__ = "MIT License"
__license_url__ = __homepage__ + "/blob/master/LICENSE"
__long_description__ = _load_readme()
__pkg_name__ = "satellighte"
__requirements__ = _load_requirements()
__version__ = pkg_version

__all__ = [
    "__author__",
    "__author_email__",
    "__copyright__",
    "__description__",
    "__homepage__",
    "__license__",
    "__license_url__",
    "__long_description__",
    "__pkg_name__",
    "__requirements__",
    "__version__",
]
